"use strict";
// ================================ Imports ================================ //
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __exportStar = (this && this.__exportStar) || function(m, exports) {
    for (var p in m) if (p !== "default" && !Object.prototype.hasOwnProperty.call(exports, p)) __createBinding(exports, m, p);
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.application = void 0;
const utilities_1 = require("./utilities");
const regexes_1 = __importDefault(require("./regexes"));
// ============================== Application ============================== //
/**
 * Application constructor
 * @param spec - Application configuration
 * @returns Application object
 */
function application(spec) {
    // Private variables
    const _helpOptionEnabled = !spec.disableHelpOption;
    const _commands = [];
    const _options = [];
    // Input parsing
    const _parseArguments = (spec, providedArgs) => {
        const args = {};
        const keys = spec.match(regexes_1.default.argumentParse);
        if (!keys)
            return args;
        const required = (() => {
            const s = keys[1].trim().replace(/[<>]/g, '');
            const a = s.length > 0 ? s.split(' ') : [];
            return a;
        })();
        const last = keys[2]?.replace(/[[\]<>.]/g, '');
        required.forEach((key, index) => {
            if (!providedArgs[index])
                throw new Error(`Missing argument: ${key}`);
            (0, utilities_1.define)(args, key, providedArgs[index]);
        });
        if (last) {
            // required.length is passed in down here
            // because it's the last index of the
            // array + 1
            if (providedArgs[required.length]) {
                (0, utilities_1.define)(args, last, keys[2]?.includes('...') // Checks if last is variadic
                    ? providedArgs.slice(required.length).join(' ').trimEnd()
                    : providedArgs[required.length]);
            }
            else if (!/[[\]]/g.test(keys[2])) {
                // Checks if last is not optional
                throw new Error(`Missing argument: ${last}`);
            }
        }
        return args;
    };
    const _parseOptions = (exec, commandOptions) => {
        const options = {};
        const stringified = exec.join(' ');
        const regexMatch = (0, utilities_1.matchAll)(regexes_1.default.optionParse, stringified);
        for (const match of regexMatch) {
            const optionKey = match[1];
            const optionMeta = [...(commandOptions || []), ..._options].find((e) => {
                return e.name === optionKey || e.alias === optionKey;
            });
            if (!optionMeta)
                continue;
            const defaultValue = optionMeta.flag ? false : optionMeta.default;
            const optionValue = optionMeta.flag || (match[2] || '').replace(/(^")|("$)/g, '');
            (0, utilities_1.define)(options, optionMeta.name, optionValue || defaultValue);
        }
        return options;
    };
    // Validation
    const _validateCommand = (command) => {
        if (command.name.length <= 0)
            throw new Error('Command names must be at least 1 character');
        if (_commands.find((e) => e.name === command.name))
            throw new Error(`Command ${command.name} already exists`);
        if (command.options)
            _validateOptions(command.options);
        if (command.arguments) {
            if (!regexes_1.default.argumentParse.test(command.arguments)) {
                throw new Error(`Arguments for command ${command.name} are formatted incorrectly`);
            }
            if (command.arguments.includes('__proto__'))
                throw new Error(`Arguments cannot be named "__proto__"`);
        }
    };
    const _validateOptions = (optionArray) => {
        for (const option of optionArray) {
            if (_options.find((e) => e.name === option.name)) {
                throw new Error(`Option ${option.name} already exists in global options`);
            }
            if (!regexes_1.default.optionValidate.test(option.name)) {
                throw new Error(`Option "${option.name}" has an incorrectly formatted name`);
            }
            if (!option.alias)
                continue;
            if (_options.find((e) => e.alias === option.alias)) {
                throw new Error(`Option ${option.name} already exists in global options, or it's alias is already in use`);
            }
            if (!regexes_1.default.aliasValidate.test(option.alias)) {
                throw new Error(`Alias "${option.alias}" for option ${option.name} is formatted incorrectly`);
            }
            if (option.name === '__proto__') {
                throw new Error(`Option "${option.name}" has illegal name`);
            }
        }
    };
    // Version option
    const _version = () => console.log(spec.version);
    _options.push({
        name: '--version',
        alias: '-v',
        description: 'Display the app version',
        flag: true
    });
    // Help option
    const _help = (command) => {
        const optionsList = [];
        [...(command?.options || []), ..._options].forEach((o) => {
            optionsList.push(`${o.name} ${o.alias || ''}`.trim());
        });
        const optionsPadLength = (0, utilities_1.maxLength)(optionsList) + 2;
        optionsList.forEach((s, i) => {
            const optionName = s.split(' ')[0];
            const optionMeta = [...(command?.options || []), ..._options].find((o) => o.name === optionName);
            optionsList[i] = (0, utilities_1.padStringTo)(s, optionsPadLength) + (optionMeta?.description || '');
        });
        if (command) {
            const usage = [];
            usage.push(command.name);
            if (command.arguments)
                usage.push(command.arguments);
            console.log(`Usage: ${usage.join(' ')}\n\n` +
                `    ${command.description}\n\n` +
                `Options:\n    ${optionsList.join('\n    ')}\n`);
        }
        else {
            const commandsList = [];
            _commands.forEach((c) => {
                commandsList.push(`${c.name} ${c.arguments || ''}`.trim());
            });
            const commandsPadLength = (0, utilities_1.maxLength)(commandsList) + 2;
            commandsList.forEach((s, i) => {
                const commandName = s.split(' ')[0];
                const commandMeta = _commands.find((c) => c.name === commandName);
                commandsList[i] = (0, utilities_1.padStringTo)(s, commandsPadLength) + (commandMeta?.description || '');
            });
            console.log(`Commands:\n    ${commandsList.join('\n    ')}\n\nOptions:\n    ${optionsList.join('\n    ')}\n`);
        }
    };
    if (_helpOptionEnabled) {
        _options.push({
            name: '--help',
            alias: '-h',
            description: 'Display help menu',
            flag: true
        });
    }
    // Public options
    const command = (command) => {
        _validateCommand(command);
        _commands.push(command);
    };
    const commands = (...commands) => {
        commands.forEach((c) => {
            command(c);
        });
    };
    const options = (...options) => {
        _validateOptions(options);
        options.forEach((e) => _options.push(e));
    };
    const run = (input = process.argv.splice(2)) => {
        const command = _commands.find((c) => c.name === input[0]);
        const options = _parseOptions(input, command?.options || []);
        if (!command) {
            if (_helpOptionEnabled)
                _help();
            return;
        }
        if (options['--version']) {
            _version();
            return;
        }
        if (options['--help'] && _helpOptionEnabled) {
            _help(command);
            return;
        }
        const args = _parseArguments(command.arguments || '', 
        // A hack to remove all options from input, due to a
        // limitation with iterating over the array instead
        (command ? input.slice(1) : input).join(' ').replace(regexes_1.default.optionParse, '').split(' '));
        command.action({ arguments: args, options: options });
    };
    // Return object
    return (0, utilities_1.deepFreeze)({
        command,
        commands,
        options,
        run
    });
}
exports.application = application;
// Re-export types for TSDoc and accessibility
__exportStar(require("./types"), exports);
